// PROJECT : ALIB
// FILE    : ALHND.H - header file for handles
// AUTHOR  : Michael Vorburger, alpha ware, CH-SG
// DATE oC : January 1994
// DATE oLM: May 1995
// LEVEL   : GLE

#ifndef __ALHND_H
#define __ALHND_H

#include <error.h>
#include <debug.h>
#include <iostream.h>

#ifndef __ALOBJECT_H
#  include <albasecl.h>
#endif

/* ------------------------------------------------------------------------ */

// FORWARD DECL
template <class T> class Shell;

/* ------------------------------------------------------------------------
    CLASS : SPtr<T> = Smart Pointer
    TYPE  : Template for ANY class T that has a "refs" data member of
				int, long or similar type and that is visible to SPtr.
            Ready-To-Use. Don't have to derrivate of SPtr<>.

    DESC  : This is a pattern for a smart pointer that frees the object
            it points to whenever neccessary. Don't delete these objects!

    REMARK: Comparison of SPtrs works like comparison of C-Pointers: They
            are equal, if they point to the same address. The objects that
            these pointers refer to are NOT compared.

    IMPL DETAIL: Have a look at the operator=() members. The order of
                 execution of these tiny commands there is important
                 if one attempts to set a SPtr to "itself". We could
                 also insert a check like if(p==r) but this would consume
                 too much space & time compared to the few cases where
                 one really performs a self-assignment. The approach
                 shown here is better and handles this special case
                 too, because the ref.count is FIRST incremented.
*/

template<class T> class SPtr {
 public:

   T* operator->() { return r;  };
   T& operator*()  { return *r; };

	const T* operator->() const { return r;  };
   const T& operator*()  const { return *r; };

   SPtr(T& o)       : r(&o)   { r->refs++;        };
   SPtr(T* p)       : r(p)    { if (r) r->refs++; };
//	SPtr(SPtr<T>& s) : r(s.r)  { if (r) r->refs++; };
	SPtr(const SPtr<T>& s) : r(s.r)  { if (r) r->refs++; };
	SPtr(Shell<T>& h); // Only decl; inline-def comes at the end.
	SPtr()           : r(NULL) {                   };

	~SPtr()
	  { if (r)
		 { r->refs--;
			if (!(r->refs)) delete r;  };
	  };

	SPtr& operator=(T& o)
     { o.refs++;
		 if (r)
		 {  r->refs--;
			 if (!(r->refs)) delete r; };
		 r=&o;
		 return *this;
	  };

	SPtr& operator=(T* p)
	  { if (p) p->refs++;
		 if (r)
		 {  r->refs--;
			 if (!(r->refs)) delete r; };
		 r=p;
		 return *this;
	  };

	SPtr& operator=(SPtr<T>& s)
	  { if (s.r) s.r->refs++;
		 if (r)
		 {  r->refs--;
			 if (!(r->refs)) delete r; };
		 r=s.r;
		 return *this;
	  };

   SPtr& operator=(Shell<T>& h); // Only decl; inline-def comes at the end.

          bool operator==(const T* p)       const { return   p==r; };
          bool operator==(const SPtr<T>& s) const { return s.r==r; };
   friend bool operator==(const T* p, const SPtr<T> SP);
          bool operator!=(const T* p)       const { return   p!=r; };
          bool operator!=(const SPtr<T>& s) const { return s.r!=r; };
   friend bool operator!=(const T* p, const SPtr<T> SP);
	// No comp with Typ T& because that could be confusing.

 public: // private: (PROBLEM)
	T* r;
};

template<class T> bool operator==(const T* p, const SPtr<T> SP)
  { return   p==SP.r; };
template<class T> bool operator!=(const T* p, const SPtr<T> SP)
  { return   p!=SP.r; };

/* ------------------------------------------------------------------------
    CLASS : Fruit
    TYPE  : ABC, belongs to my Shell/Fruit-Pattern
    DESC  : This class has to be mixed-in in any other class that will be
            used as a FRUIT in connection with the Shell<> template.

    PROBLEMS: 1. Security problem in connection with templates. The "refs"
             member and the destructor should be private. I want to make
             ALL Shell<> and all SPtr<> templates friends of this class
				 Fruit. How can this be set??

              2. Static fruits are attempted to be delete'd at the end of
             the program. This is a mistake because they were never new'd.
             The problem could be solved if....
             ...but there is no known possibility to distinguish from an
             object wheter its dynamical or statical. I tried to find this
             out by checking at the beginning if new() was invoked, but it
             doesn't work this way...

// Fruit nur als "Frucht" fr alShellTempl verwenden!
// ALL Fruit-Objects MUST be dynamic, NOT static!
// NEVER delete an Fruit!

// MAKRO  : Fr jede Fruit-Klasse K muss genau 1x das FRUITMACRO im
// public-Teil auftauchen. Siehe Bemerkung unten. Setzt einen
// Copy-Konstruktor (ev. autom. gen.) voraus.
// Gilt _NICHT_ fr ABCs.

    IMPL DETAIL: 1. The Copy() member function: This function simply
                    invokes the copy-constructor of the class. This is
                    neccessary because std copy-constructors cannot be
                    virtual (but this member can).
*/
#define FRUITMACRO(F) Fruit& Copy() const { return ref_cast(F,*new(F(*this))); };
/*
#define FRUITMACRO(F) \
   public:    Fruit& Copy() const { return ref_cast(F,*new(F(*this))); }; \
   private: void operator delete(void*);                                \
   public:
*/
// The new C++ std draft will allow the following (instead the ref_cast):
//      F& Copy() const { return *new F(*this)); };


class Fruit : public AlBObject
{public:
	Fruit()               : refs(0) {};
   Fruit(const Fruit& C) : refs(0) {}; // _NICHT_ refs(C.refs) !!!

   void MakeStatic() const     { const_cast(Fruit* const, this)->refs=-2; };
/* Erklrung: Fruits MSSEN eigt. immer dynamische Heap-Objekte sein. In
   Ausnahmefllen ist es aber erlaubt, sie statisch zu machen. Dann muss aber
   diese Methode fr das Objekt aufgerufen werden.
   Diese Sache ist noch nicht vollstndig & sicher berdacht... ev. mssen
   solche Fruits immer const sein??                                         */

 protected:
   virtual Fruit& Copy() const = 0;  // Siehe Makro oben.

 public: // Should be <private:> See PROBLEM above.

   SWORD refs;      // How many references to this object?
                    // special cases: -1 = dead object, -2 = static object
	virtual ~Fruit()
     { if ((refs--)>0)
         Error(ErrWrongDelete);
       refs=-1;                };

   // virtual??? MUST BE! - NO, see trick in Shell::operator==();
   bool operator==(const Fruit& C) const; // This func has NO DEFINITION. That's
   // ok. If a user wants to compare Hnd (say: also Fruits) he must redefine an
   // operator==(const DerrivedFruitType&) later. If a user compares
	// Hnds<DerrF> without redefining DerrF::operator==(DerrF), the
   // _linker_ generates an error.

   friend class ShellEngine;
// Die folgenden FRIEND Deklarationen bewirkt _nix_ . Siehe oben.
// friend class SPtr<Fruit>;
// friend class Shell<Fruit>;
   // darum weglassen; No, bool Shell::operator==(const Shell<T>& c) const
   // geht sonst nicht.

// RELICT
//   virtual BOOL IsA(UWORD id) const  { return( id == IdFruit ); };
//   virtual UWORD TypeOf()     const  { return IdFruit;          };
};


/* ------------------------------------------------------------------------
    MACROS: DERR_SHELL*
    DESC  : One or more of these macros has to be used in every class S that
            is derrived from a Shell<F> named P.
    REMARK: Consider using one of the macros in the ALIAS section below.
*/
// SHELL_INTERFACE(self.Typ, FruitTyp) genau 1x pro Shell.
#define DERR_SHELL_INTERFACE_CSTR(S,P,F) \
        S()                        : P()    { }; \
        S(const F& arg)            : P(arg) { }; \
		  S(const Shell<F>& arg)     : P(arg) { }; \
	 bool operator==(const S& c) const            \
       { if(ShellEngine::operator==(c)) return true;       \
         else return ref_cast(const F,FrtConst())==ref_cast(const F,c.FrtConst()); }; \
    bool operator!=(const S& c) const            \
       { if(ShellEngine::operator==(c)) return false;      \
         else return !( ref_cast(const F,FrtConst())==ref_cast(const F,c.FrtConst()) ); };
//        S(const F* arg)            : P(arg) { }; \
//        S(const Shell<F>* arg)     : P(arg) { }; \

#define DERR_SHELL_INTERFACE_ASSG(S,P,F) \
        P& operator=(const F& arg) { EqualFrtPtr(&arg); return *this; }; \
        P& operator=(const S& arg) { EqualShell(arg); return *this;}; //\
//        Shell<F>& operator=(const ShellEngine& S)  { ShellEngine::EqualShell(S); return *this; };
//        Shell<F>& operator=(const F* arg) { return EqualFruitPtr(arg);  }; \
//        Shell<F>& operator=(const Shell<F>* arg) { return EqualShell(*arg);    }; //\

#define DERR_SHELL_INTERFACE(S,P,F) DERR_SHELL_INTERFACE_CSTR(S,P,F) \
                                    DERR_SHELL_INTERFACE_ASSG(S,P,F)

/* ------------------------------------------------------------------------
    MACROS: DERR_SHELL*
    DESC  : One or more of these macros
    REMARK: Consider using one of the macros in the ALIAS section below.
*/
// FRUIT_INTERFACE(self.Typ, FruitTyp, InitializerTyp) bel. viele Male. Die
// Shell kann dann mit InitializerTyp initialisiert werden. Achtung: Die
// Entsprechende Fruit braucht einen Konstruktor UND EINEN operator=() [!!]
// mit Arg InitTyp.
#define DERRSH_FRTINI_INTF_FAM(S,SP,F,FP,T) \
        S(const T  init)                   : SP( *new F(init)  )       {  }; \
        Shell<FP>& operator=(const T  arg) { return EqualFrtPtr(new F(arg)); };/* \
        S(const T* init)                   : SP( new F(*init)  )      {  };  \
        Shell<FP>& operator=(const T* arg) { return EqualFruitPtr(new F(*arg));};
*/
#define DEC_DERRSH_FRTINI_INTF_FAM(S,SP,F,FP,T) \
        S(const T  init)                   ; \
		  Shell<FP>& operator=(const T  arg) ; \
        S(const T* init)                   ; \
        Shell<FP>& operator=(const T* arg)

#define DEF_DERRSH_FRTINI_INTF_FAM(S,SP,F,FP,T) \
        S::S(const T  init)                : SP( *new F(init)  )       {  }; \
        Shell<FP>& S::operator=(const T  arg) { return EqualFrtPtr(new F(arg)); }; //\

#define FRUITINIT_INTERFACE_FAM(S,F,FP,T) \
        DERRSH_FRTINI_INTF_FAM(S,Shell<FP>,F,FP,T)
#define DEC_FRUITINIT_INTERFACE_FAM(S,F,FP,T) \
        DEC_DERRSH_FRTINI_INTF_FAM(S,Shell<FP>,F,FP,T)
#define DEF_FRUITINIT_INTERFACE_FAM(S,F,FP,T) \
        DEF_DERRSH_FRTINI_INTF_FAM(S,Shell<FP>,F,FP,T)

#define DERRSHELL_FRUITINIT_INTERFACE(S,SP,F,T) \
        DERRSH_FRTINI_INTF_FAM(S,SP,F,F,T)
#define DEC_DERRSHELL_FRUITINIT_INTERFACE(S,SP,F,T) \
		  DEC_DERRSH_FRTINI_INTF_FAM(S,SP,F,F,T)
#define DEF_DERRSHELL_FRUITINIT_INTERFACE(S,SP,F,T) \
        DEF_DERRSH_FRTINI_INTF_FAM(S,SP,F,F,T)

#define FRUITINIT_INTERFACE(S,F,T) \
        FRUITINIT_INTERFACE_FAM(S,F,F,T)
#define FRUITINIT_INTERFACE_DEC(S,F,T) \
        DEC_DERRSH_FRTINI_INTF_FAM(S,F,F,T)
#define FRUITINIT_INTERFACE_DEF(S,F,T) \
        DEF_FRUITINIT_INTERFACE_FAM(S,F,F,T)
/*
// If F cannot be defined at the moment of declaration, use this:
#define FRUITINIT_INTERFACE_FAM_DEC(S,F,P,T) \
        S(const T  init); \
        Shell<P>& operator=(const T  arg); //\
//        S(const T* init); \
//        Shell<P>& operator=(const T* arg);
#define FRUITINIT_INTERFACE_FAM_DEF(S,F,P,T) \
		  S::S(const T  init)                         : Shell<P>( *new F(init)  )    {  }; \
        Shell<P>& S::operator=(const T  arg) { return EqualFrtPtr(new F(arg)); }; //\
//        S::S(const T* init)                  { EqualFruitPtr(new F(*init));}; \
//        Shell<P>& S::operator=(const T* arg) { return EqualFruitPtr(new F(*arg));};

// Diese Makros braucht es, weil Konstruktoren und operator= nicht "richtig"
// vererbt werden.

*/
/* ------------------------------------------------------------------------
    CLASS : ShellEngine
    TYPE  : For special purposes only! NOT DESIGNED FOR END USERS.
            You have to fully understand the Shell/Fruit pattern to be able
            to use this class.

    DESC  : This is kind of a "general shell" for all fruits. It's main purpose
            is for memory optimization; eg. in the Container class library.

	 REMARK: Comparison of ShellEngine works like comparison of two static
            objects: They are equal, if either the _internal_ pointer points
            to the same object OR if the objects _internally_ refered to REALLY
            ARE the same. This is determined by invoking operator==() on the Fruits.

    PROBLEM: Einig Methoden von Shell sollten ein Err-Objekt zurckliefern
             knnen. Idee: template <class T, T ErrT> Shell<TermFruit, ErrTerm>
             Das geht leider nicht; obwohl zB template <class T, int sz> erlaubt
             ist. Kommt das noch im bernchsten ANSI-Draft?
*/

class ShellEngine
{ public:
	 operator bool() const
      { if (r!=NULL) return FrtConst();
        else         return false;      };

    ShellEngine()                                                    { };
	 ShellEngine(const ShellEngine& arg)  : r(arg.r)                  { };
    ShellEngine(const ShellEngine* arg)  : r(arg->r)                 { };
    ShellEngine(const Fruit& arg)        : r(const_cast(Fruit&,arg)) { };
    ShellEngine(const Fruit* arg)        : r(const_cast(Fruit*,arg)) { };
    DERR_SHELL_INTERFACE_ASSG(ShellEngine,ShellEngine,Fruit)

  protected:
    const Fruit& FrtConst() const
    { if (r==NULL)
        { Error(ErrNotInitVar); /* return ErrT; s.u. */ };
      // else
        return *r;  };

    Fruit& FrtUpd()
      { Fruit& newFrt=FrtCopy();
        EqualFrtPtr(&newFrt);
        return newFrt;              };

	 Fruit& FrtMustCpy() const // MyFrtMustCopy
    { if (r==NULL)
        { Error(ErrNotInitVar); /* return ErrT; s.u. */ };
      //else
          return ref_cast(Fruit,r->Copy());
          // FALSCH wre: *new Fruit(*r); weil so Probleme mit verberbten Fruit's
          // ref_cast wg. prob bei ndern von Copy() return type (so.)
    };

    // Cannot remember why I made this protected???
    bool operator==(const ShellEngine& c) const
      { if(r==c.r) return true;
        else       return (*r)==c; };

    ShellEngine& EqualFrtPtr(const Fruit* arg)
    { if (arg!=r) r=const_cast(Fruit*,arg);
      return *this;                                                  };

	 ShellEngine& EqualShell(const ShellEngine& arg)
    { if(arg.r!=r) r=const_cast(SPtr<Fruit>&,arg.r);
      return *this;                                                  };

//  private:
public:
    SPtr<Fruit> r; // former: Fruit* r;

    Fruit& FrtCopy() const
    { if (r==NULL)
        { Error(ErrNotInitVar); /* return ErrT; s.u. */ };
      // else {
        if(r->refs>1 || r->refs==-2)
          return ref_cast(Fruit,r->Copy()); // so.
        else
          return const_cast(Fruit&,*r); /* }; */                                        };

 // a) friend template <class T> class Shell;
 // b) friend class Shell;
 // c) friend Shell<T>;
 //friend class Shell<Fruit>;
};

/* ------------------------------------------------------------------------
    CLASS : Shell<T>
    TYPE  : Main template belonging to my Shell/Fruit-Pattern.

    DESC  : Read the corresponding sections in my sjf-report.

    REMs  : T must be derrived from class Fruit!
*/

template <class T>
class Shell : public ShellEngine
{   typedef ShellEngine E;
  public:
/*
    operator bool() const   { return E::operator bool(); };
    bool operator==(const Shell<T>& c) const
       { if(E::operator==(c)) return true;
         else return ref_cast(const T,FrtConst())==ref_cast(const T,c.FrtConst()); };
    bool operator!=(const Shell<T>& c) const
       { if(E::operator==(c)) return false;
         else return !( ref_cast(const T,FrtConst())==ref_cast(const T,c.FrtConst()) ); };
*/

   DERR_SHELL_INTERFACE_CSTR(Shell,ShellEngine,T)
   Shell(const ShellEngine& se): ShellEngine(se)  {};
   DERR_SHELL_INTERFACE_ASSG (Shell,Shell,T)
/* These methodes (constructors and operator=) _have to_ be repeated here, because
    a) one might use type Shell<Fruit> directly _without_ derriving from it
        (though that's what ShellEngine is thought for... ;-)
    b) it is not possible to call the ShellEngine cstr from a class derrived from
       Shell<> (though this class also IS-A ShellEngine)
*/
  protected:
    T& Frt()                { return ref_cast(T,E::FrtUpd());         };
    const T& Frt()  const   { return ref_cast(const T,E::FrtConst()); };

    Shell<T>& EqualFrtPtr(const T* arg)    { E::EqualFrtPtr(arg); return *this;  };
    Shell<T>& EqualShell(const Shell& arg) { E::EqualShell(arg);  return *this;  };
    T&        FrtMustCpy() const           { return ref_cast(T,E::FrtMustCpy()); };

// the following is for backward compatibility, better use the substitutes now
    const T& MyFrtConst() const           { return Frt(); };
    Shell<T>& EqualFruitPtr(const T* arg) { return EqualFrtPtr(arg); };

// RELICT
// protected:
//  virtual BOOL IsA(UWORD id) const    { return( id == IdShell ) };
//  virtual UWORD TypeOf()     const    { return IdShell;         };
};

/* ------------------------------------------------------------------------
    Out-Of-Class Definitions

    DESC: These member functions could not be defined above (only declared).
          This is because the Compiler didn't know some important things at
          the declaration point.

          The functions are still inline, though.
*/

template<class T>
inline SPtr<T>::SPtr(Shell<T>& h) : r(ref_cast(SPtr<T>,h.r).r)  { if (r) r->refs++; };
	/* REM about the complicated cast:
                                     SPtr<Fruit> s=h.r;
												 SPtr<T>& p=ref_cast(SPtr<T>,h.r);
                                     r=ref_cast(SPtr<T>,h.r).r;             */
template<class T>
inline SPtr<T>& SPtr<T>::operator=(Shell<T>& h)
  { T* p=ref_cast(SPtr<T>,h.r).r;
    if (p) p->refs++;
	 if (r)
	 { r->refs--;
		if (!(r->refs)) delete r; };
	 r=p;
	 return *this;
  };


/* ------------------------------------------------------------------------
	ALIASES

		Hnd<T>               = Shell<T>
		SHELL_INTERFACE(S,F) = DERR_SHELL_INTERFACE(S,Shell<F>,F)
		SHELL_INTF(S,F)      = SHELL_INTERFACE(S,F)

	PROBLEM: I'm sorry I couldn't replace these macros with typedefs, but this
				didn't work:
								 typedef Shell Hnd;
					 ???      typedef template <class T> Shell<T> Hnd<T>;
								 template <class T> typedef ...
								 ...
*/
#define Hnd Shell

#define SHELL_INTERFACE(S,F) DERR_SHELL_INTERFACE(S,Shell<F>,F)
#define SHELL_INTF(S,F) DERR_SHELL_INTERFACE(S,Shell<F>,F)

/* ------------------------------------------------------------------------
	 MACRO: CHECK_SHELL(YourDerrivedShellClass)
	 DESC:  You might remember that a Shell MUST have the same size as a
			  native pointer or as a ShellEngine object. This macro chechks
			  that at compile time.

	 PROBLEM: Does not work! It seems it's not possible to include
				 preprocessor commands in a #define command because the #
				 is always treated as token. Know a solution?
*/
/*
#define CHECK_SHELL \
	#if (sizeof(S)!=sizeof(ShellEngine))                                 \
	#  error This handle has additional data members or virtual members  \
	#endif
*/
// ------------------------------------------------------------------------

#endif  // __ALHND_H
