// PROJECT : ALIB (alpha library)
// FILE    : LIST.H - header file for Container-Class library
// AUTHOR  : Michael Vorburger, alpha ware, CH-SG
// DATE oC : September 1994
// DATE oLM: October 1994
// LEVEL   : GLE

/* DESCRIPTION

   This module provides a very flexible CONTAINER family with Stacks & Queues
   and Orderer COLLECTIONS such as Arrays, Lists etc.
   Interface (declaration, ABCs) and Implementation are strictly separated.
*/


//#include <debug.h>
#include <portab.h>
#include <error.h>
#include <alobject.h>
#include <alstring.h>

#pragma warn -inl

/* ------------------------------------------------------------------------ */
// FORWARD-Deklarationen
template<class T> class ArrayABC;
template<class T> class AIterABC;

/* ------------------------------------------------------------------------ */
// ContainerABC<I, O> - abstract base class for EVERYTHING here
// methods: Add(I)    - add item to Container
//          GetRemove - return any item (not known which!) and remove it
//          IsEmpty   - is the Container empty?
template<class I, class O> class ContainerABC
{ public:
   virtual void        Add(const I&)   = 0;
   virtual O           GetRemove()     = 0;
   virtual const bool  IsEmpty()       = 0;
   virtual             ~ContainerABC()    { };
// short forms
   ContainerABC& operator+=(const I& e)   { Add(e); return *this; };
   ContainerABC& operator<<(const I& e)   { Add(e); return *this; };
   O             operator() ()            { return GetRemove();   };
                 operator const bool()    { return !IsEmpty();    };

   ContainerABC() { };
  private:
    operator=(const ContainerABC&); ContainerABC(const ContainerABC&);
};

/* ------------------------------------------------------------------------ */
// StackABC<T> - interface for Stack classes
// A Stack is a stack. If you don't know what this is -> :-(
// (it's a Container that works after the LIFO principle)
template<class T> class StackABC : public ContainerABC<T, T>
{ public:
};

// StackT2_E<Type,BaseType,Implementation> - type conversion template for
// a <Type> stack that it is implemented with a <Implementation> of <BaseType>
template<class T, class BT, class Eng> class  StackT2_E : public StackABC<T>
{   Eng E; // the "engine" or implementation
 public:
   void        Add(const T& a)   { E.Add(a);                };
   T           GetRemove()       { return (T)E.GetRemove(); };
   const bool  IsEmpty()         { return E.IsEmpty();      };
};
// StackP_E<T,Implementation> - a stack for Pointers. Implementation must be
// of Type <void*>. This template bases on StackT2_E.
template<class T, class E> class StackP_E : public StackT2_E<T,void*, E> {};
// StackP<T> a stack of Pointers. Implemented with Stack_LL.
template<class T> class StackP : public StackT2_E<T,void*, Stack_LL<void*> > {};
// StackH<T> a stack of Handles. Implemented with Stack_LL.
template<class T> class StackH : public StackT2_E<T,ShellEngine, Stack_LL<ShellEngine> > {};


// Stack_LL<T> - an implementation of a stack. works with a single linked list.
template<class T> class Stack_LL : public StackABC<T>
{ public:
   void       Add(const T& a);
   T          GetRemove();
   const bool IsEmpty()        { return last==NIL; };

   Stack_LL() : last(0), head(0)  { };
   ~Stack_LL();

  private:
    Node<T>* last;
    Node<T>* head;
};
// Because this is the only implementation avalaible at the moment, we define
// Stack as Stack_LL.
#define StackTpl Stack_LL

/* ------------------------------------------------------------------------ */
// QueueABC<T> - interface for Queue classes
// A Queue is a Container that works after the FIFO principle.
template<class T> class QueueABC : public ContainerABC<T, T>
{ public:
};


// QueueT2_E<Type,BaseType,Implementation> - type conversion template for
// a <Type> stack that it is implemented with a <Implementation> of <BaseType>
template<class T, class BT, class Eng> class  QueueT2_E : public QueueABC<T>
{   Eng E; // the "engine" or implementation
 public:
   void        Add(const T& a)   { E.Add(a);                };
   T           GetRemove()       { return (T)E.GetRemove(); };
   const bool  IsEmpty()         { return E.IsEmpty();      };
};
// QueueP_E<T,Implementation> - a Queue of Pointers. Implementation must be
// of Type <void*>. This template bases on QueueT2_E.
template<class T, class E> class QueueP_E : public QueueT2_E<T,void*, E> {};
// QueueP<T> a Queue of Pointers. Implemented with Stack_LL.
template<class T> class QueueP : public QueueT2_E<T,void*, Queue_LL<void*> > {};
// QueueH<T> a stack of Handles. Implemented with Stack_LL.
template<class T> class QueueH : public QueueT2_E<T,ShellEngine, Queue_LL<ShellEngine> > {};


// Queue_LL<T> - an implementation of a queue; works with a single linked list
template<class T> class Queue_LL : public ContainerABC<T, T>
{ public:
   void       Add(const T& a);
   T          GetRemove();
   const bool IsEmpty()        { return head==NIL; };

   Queue_LL() : last(0), head(0)  { };
   ~Queue_LL();

  private:
    Node<T>* last;
    Node<T>* head;
};
template<class T> class QueueTpl : public Queue_LL<T> {};

/* ------------------------------------------------------------------------ */
// CollectionABC<I, O> & IterABC<I, O> - abstract base class for special
// Container where EACH item can be accessed (and found and removed)
// This work with ITERATORs: You can Find(I) an item and you'll get an Iter
// With thisone, you can and Remove(Iter) the item.

template<class I, class O> class IterABC;
template<class I, class O> class Iter;

template<class I, class O> class CollectionABC : public ContainerABC<I,O>
{  typedef IterABC<I,O> It;
 public:
   virtual O           Get(const It& i)   const        { return i.Get(); };
   virtual Iter<I,O>   Find(const I&)              const  = 0;
   virtual void        Remove(const It&)                  = 0;
};

template<class I, class O> class IterABC : public Fruit
{   operator=(const IterABC&); IterABC(const IterABC&);
  public:
    virtual bool     Valid()                                const = 0;
    virtual bool     ValidFor(const CollectionABC<I, O>& c) const = 0;
    virtual O        Get()                                  const = 0;

    operator bool()                            { return Valid(); };
    virtual O  operator() ()                   { return Get();   };

    IterABC() { };
};

template<class I, class O> class Iter : public Hnd<IterABC<I,O> >
{   typedef IterABC<I,O> F; typedef CollectionABC<I,O> C;
 public:
    SHELL_INTERFACE(Iter,F)

    bool     Valid() const                { return Frt().Valid(); };
    bool     ValidFor(const C& c) const   { return Frt().ValidFor(c); };
    O        Get()                const   { return Frt().Get(); };

    operator bool() const                 { return Valid(); };
    O operator() () const                 { return Get();   };

    operator const F&() const             { return Frt(); };
};

/* ------------------------------------------------------------------------ */
// Set<T>
// A Set is an unordered set (:-) of items; so you cannot iterate (++)
// through it!
template<class T> class Set : public CollectionABC<T, T>
{ public:
}; // Q: Does this make sense??

/* ------------------------------------------------------------------------ */
// OCollectionABC<I, O> & OIterABC<I, O> - abstract base class
// An OrdCollection is an ordered thing: You can access the "third" item and
// you can iterate (by using UWORDs or OIters) through it.
// Attention: The order _always_ starts from one (1) on and goes up to a
// fixed or variable size which can be known be calling the Items() method.
// It is NOT allowed to derrive any class that allowes a range like [-5..3].
// new methods are: - Get and Remove with UWORDS instead Iterators
//                  - bool Items() to know how many items are avail.
//                  - NewIter() to get a new OIter. This is very usefull.

template<class I, class O> class OIterABC;
template<class I, class O> class OIter;

template<class I, class O> class OCollectionABC : public CollectionABC<I, O>
{  typedef IterABC<I,O> It;
 public:
   virtual O           Get(const UWORD)           const ;
   virtual void        Remove(const UWORD r)            ;
   virtual UWORD       Items()                    const = 0;
   virtual Iter<I, O>  NewIter()                  const { return NewOIter(); };
   virtual OIter<I,O>  NewOIter()                 const = 0;
// inherited:
   O                   GetRemove()                      ;
   const bool          IsEmpty()                        { return Items()==0; };
   virtual O           Get(const IterABC<I,O>& i) const { return i.Get();    };
   virtual Iter<I, O>  Find(const I& f)           const { return Find_OIter(f);    };
   virtual OIter<I,O>  Find_OIter(const I&)       const = 0;

   O                   operator[](UWORD i)        const { return Get(i);  };
   O                   operator[](const It& i)    const { return i.Get(); };
};

template<class I, class O> class OIterABC : public IterABC<I, O>
{   operator=(const OIterABC&); OIterABC(const OIterABC&);
  public:
    virtual OIterABC& operator++ () = 0;
    virtual OIterABC& operator-- () = 0;
    virtual void   Set(UWORD p)     = 0;
    //virtual void operator++ (int) = 0;
    //void operator-- (int)         = 0;

    OIterABC() { };
};

template<class I, class O> class OIterErr : public OIterABC<I,O>
{ public:
    bool     Valid()                                const { return false; };
    bool     ValidFor(const CollectionABC<I, O>& c) const { return false; };
    O        Get()                                  const { Error(ErrWrongIter); };

    OIterABC<I,O>& operator++()                            { Error(ErrWrongIter); };
    OIterABC<I,O>& operator-- ()                           { Error(ErrWrongIter); };
    void           Set(UWORD p)                            { Error(ErrWrongIter); };

    OIterErr() {};
    OIterErr(const OIterErr& c) { };
    typedef OIterErr<I,O> td; FRUITMACRO(td)
};

template<class I, class O> class OIter : Hnd<OIterABC<I,O> >
{ public:
    typedef OIterABC<I,O> F; SHELL_INTERFACE(OIter,F)

    OIter& operator++ ()  { Frt().operator++(); return *this;            };
    OIter& operator-- ()  { Frt().operator--(); return *this;            };
    OIter  operator++(int){ OIter<I,O> O=*this; Frt().operator++(); return O; };
    OIter  operator--(int){ OIter<I,O> O=*this; Frt().operator--(); return O; };
    void   Set(UWORD p)   { Frt().Set(p);              };

    bool   Valid() const  { return Frt().Valid(); };
    bool   ValidFor(const CollectionABC<I, O>& c) const { return Frt().ValidFor(c); };
    O      Get() const    { return Frt().Get(); };

    operator bool() const              { return Valid(); };
    O operator() () const              { return Get();   };

    operator const Iter<I,O>()      const { return Frt(); };
    operator const IterABC<I,O>&()  const { return Frt(); };
    operator const OIterABC<I,O>&() const { return Frt(); };
};

/* ------------------------------------------------------------------------ */
// ArrayABC<T> - interface for Array classes
// An Array behaves like a built-in array: It's size is fixed.
template<class T> class ArrayABC : public OCollectionABC<T,T>
{ public:
};

//template<class T> class List_A; // forward declaration for friend below. GEHT NICHT!

// Array_A<T> - an implementation of an Array. Works internally with an array
//                                                  (how surprising :-))
template<class T> class Array_A : public ArrayABC<T>
{ public:
   void       Add(const T& i)                   ;
   T          Get(const UWORD i)          const { CheckBnd(i); return arr[i-1];   };
   OIter<T,T> Find_OIter(const T&)        const ;
   void       Remove(const UWORD r)             ;
   void       Remove(const IterABC<T,T>&)       ;
   OIter<T,T> NewOIter()                  const { return *new(AIter_A<T>(*this)); };
   UWORD      Items()                     const { return (n-1); /* s; ?? */       };

               Array_A(UWORD s);
               Array_A(const Array_A<T>& cpy);
   Array_A<T>& operator= (const Array_A&);
               ~Array_A()                       { delete[] arr; };

//   friend class List_A<T>;
  public: // protected: (Prob: obenstehender Friend funktioniert nicht)
    UWORD s;    // size, valid Array indizes are in [1..s]
    UWORD n;    // next item position for a new item, valid range is [1..s]
    T*    arr;  // ptr to real C-array
    bool ValidBnd(UWORD j) const { return((j>0) && (j<n) ); };
    void CheckBnd(UWORD j) const { if(!ValidBnd(j)) Error(ErrArrayBounds); };
  friend class AIter_A<T>;
};
#define ArrayTpl  Array_A // typedef doesn't work with templates
#define ArrayP(P) Array_A<P>
#define ArrayH(H) Array_A<H>

template<class T> class AIter_A : public OIterABC<T,T> // AIterABC<T>
{   void Checki() const { C.CheckBnd(i); };
  public:
    bool       Valid() const     { return C.ValidBnd(i);               };
    bool       ValidFor(const CollectionABC<T, T>& c) const { if(&C!=&c) Error(ErrWrongIter); /*else*/ return true; };
    T          Get() const       { return C.Get(i); };

    OIterABC<T,T>& operator++ () { Checki(); ++i; return *this;        };
    OIterABC<T,T>& operator-- () { Checki(); i--; return *this;        };
    void        Set(UWORD p)     { C.CheckBnd(p); i=p;                 };

    AIter_A(const Array_A<T>& c)          : C(c) { i=1;                  };
    AIter_A(const Array_A<T>& c, UWORD p) : C(c) { C.CheckBnd(p); i=p;   };
    AIter_A(const AIter_A& s)             : C(s.C), i(s.i)             { };
    AIter_A& operator=(const AIter_A& s)  { ValidFor(s.C); i=s.i; /*s.i=0;
                                             delete &s;*/ return *this;    };
    FRUITMACRO(AIter_A<T>)
  public: // PROTECTED: (Prob: Untenstehender Friend funktioniert nicht)
    const Array_A<T>& C;
    UWORD i;                  // currently "iterated" item; range [1..C.s]
  friend class Array_A<T>;
  //friend void List_A<T>::Insert(const IterABC<T,T>& I, const T it);
};
#define AIter AIter_A // typedef doesn't work with templates
/*
// ArrayT2_E<Type,BaseType,Implementation> - type conversion template for
// a <Type> Array that it is implemented with a <Implementation> of <BaseType>
template<class Tx, class BT, class Eng> class ArrayT2_E : public ArrayABC<Tx>
{ / Eng E; // the "engine" or implementation
    typedef OIter<T,T> It;
 public:
   void       Add(const T& i)                   { E.Add(i);               };
   T          Get(const UWORD i)          const { return E.Get(i);        };
   OIter<T,T> Find_OIter(const T& f)      const { return (It)E.Find_OIter(f); };
   void       Remove(const IterABC<T,T>& i)     { E.Remove(i);            };
   OIter<T,T> NewOIter()                  const { return (It)E.NewOIter();    };
   UWORD      Items()                     const { return E.Items();       };

              ArrayT2_E(UWORD s)            :  E(s) { };
  ArrayT2_E<T,BT,Eng>& operator= (const Eng& a) { E.operator=(a); return *this; };
};

// ArrayH<T> an array of Handles. Implemented with Array_A.
template<class T> class ArrayH
  : public ArrayT2_E<T,ShellEngine,Array_A<ShellEngine> >
  { ArrayH(UWORD s) : ArrayT2_E<T,ShellEngine,Array_A<ShellEngine> >(s) { };
  };
/*
// ArrayP_E<T,Implementation> - an Array of Pointers. Implementation must be
// of Type <void*>. This template bases on ArrayT2_E.
template<class T, class E> class ArrayP_E : public ArrayT2_E<T,void*, E> {};
// ArrayP<T> an array of Pointers. Implemented with Array_A.
template<class T> class ArrayP : public ArrayT2_E<T,void*,Array_A<void*> > {};
*/
*/
/* ------------------------------------------------------------------------ */
// List<T> & LIter<T>
// Lists are similar to Arrays, but they are "unlimited" - they never exceed.
// Lists have an additional method Insert() which does not override any item.

// PROBLEM: Is List a child of Array() ? It has the same interface; but the
// meaning changed a little... I don't know yet. I'll assume it is... NO!
// this multiple inheritance causes troubles because double OColl is amgib!
template<class T> class ListABC : public OCollectionABC<T,T> // ? ArrayABC<T>
{public:
   virtual void Insert(const IterABC<T,T>&, const T) = 0;
   virtual void Insert(const UWORD r, const T it)    = 0;
 //    { OIter<T,T> I=NewOIter(); I.Set(r); Insert(I,it); };
 // Die Deklaration von diesem Insert muss sowieso wiederholt werden....
};

// List_A - a list implementation with a flexible array
template<class T> class List_A : public ListABC<T> //, private Array_A<T>
{  //typedef Array_A<T> A;
 public:
   void       Add(const T& i)                   ;
   void       Insert(const IterABC<T,T>&, const T);
   void Insert(const UWORD r, const T it)
     { OIter<T,T> I=NewOIter(); I.Set(r); Insert(I,it); };
   T          Get(const UWORD i)          const { return A.Get(i);        };
   OIter<T,T> Find_OIter(const T& i)      const { return A.Find_OIter(i); };
   void       Remove(const UWORD r)             { A.Remove(r);            };
   void       Remove(const IterABC<T,T>& i)     { A.Remove(i);            };
   OIter<T,T> NewOIter()                  const { return A.NewOIter();    };
   UWORD      Items()                     const { return A.Items();       };

   List_A(UWORD iniSz=10, UWORD icr=5) : A(iniSz), incr(icr)            { };
   List_A(const List_A& copyL)         : A(copyL.A), incr(copyL.incr)   { };
   List_A& operator= (const List_A& copyL) { A.operator=(copyL.A); incr=copyL.incr; };

/* This MUST BE REPEATED here if private derr of Array_A AND ListABC
   ContainerABC<T,T>& operator+=(const T& e)   { Add(e); return *this;  };
   ContainerABC<T,T>& operator<<(const T& e)   { Add(e); return *this;  };
   T                  operator() ()            { return A.GetRemove(); };
                      operator const bool()    { return !A.IsEmpty();     };
*/
 protected:
    Array_A<T> A;
    UWORD incr; // um wieviel Elemente wird size bei Bedarf erhht?
};
#define ListTpl List_A
#define ListP(P) List_A<P>
#define ListH(H) List_A<H>

/* ------------------------------------------------------------------------ */
// SortListABC<T>
// SortLists are similar to Lists, but all elements are automatically sorted
// Take care: You have no idea & control at which position an item is stored;
// but SortLists are still OrdCollections because you can iterate through
// them. A SortList CAN contain an item twice or more.
// REMARK: Though SortList has the same interface as List, it is not a List
// in the OOD way (in my opinion) because it behaves quite different; and it
// has no insert method.
template<class T> class SortListABC : public OCollectionABC<T, T>
{public:
};

// SortList_A - a sorted list implemented with List_A
template<class T> class SortList_A : public SortListABC<T>
{ public:
   void       Add(const T& i)                   ;
   T          Get(const UWORD i)          const { return A.Get(i);        };
   OIter<T,T> Find_OIter(const T& i)      const { return A.Find_OIter(i); };
   void       Remove(const UWORD r)             { A.Remove(r);            };
   void       Remove(const IterABC<T,T>& i)     { A.Remove(i);            };
   OIter<T,T> NewOIter()                  const { return A.NewOIter();    };
   UWORD      Items()                     const { return A.Items();       };

   SortList_A(UWORD iniSz=10, UWORD icr=5) : A(iniSz,icr)               { };
   SortList_A(const SortList_A& copyL)     : A(copyL.A)                 { };
   SortList_A& operator= (const SortList_A& copyL){ A.operator=(copyL.A); };

 protected:
   List_A<T> A;
};
#define SortListTpl SortList_A
#define SortListP(P) SortList_A<P>
#define SortListH(H) SortList_A<H>

/* SortList_LL - a sorted list implemented with a double linked list
   DOES NOT WORK! PLEASE FINISH LATER!
*/
template<class T> class SortList_LL : public OCollectionABC<T,T>
{ };
/* typedef IterABC<T,T> It;
 public:
   void       Add(const T& i);
   T          Get(const It& i)            const { return i.Get();         };
   T          Get(const UWORD i)          const {};
   OIter<T,T> Find_OIter(const T& i)      const {};
   void       Remove(const It& i)         {};
   OIter<T,T> NewOIter()                  const { return *new SIter_LL<T>(*this); };
   UWORD      Items()                     const { return l;};

   SortList_LL() : h(NULL), l(0) {};
   SortList_LL(const SortList_LL& copyL);
   SortList_LL& operator= (const SortList_LL& copyL);
   ~SortList_LL();
   friend class SIter_LL<T>;
  private:
     DNode<T>* h;
     UWORD l;
};
template<class T> class SIter_LL : public OIterABC<T,T>
{   //operator=(const OIterABC&); OIterABC(const OIterABC&);
  public:
    bool     Valid()                               const { return i!=NULL; };
    bool     ValidFor(const CollectionABC<T,T>& c) const { if(&C!=&c) Error(ErrWrongIter); /*else*/ return true; };
    T        Get()                                 const { CheckI(); return i->item; };

    OIterABC<T,T>& operator++ ()   { CheckI(); if (i) i=i->n; else /*###*/return *this; };
    OIterABC<T,T>& operator-- ()   { CheckI(); if (i) i=i->p; else /*****/return *this; };
    void   Set(UWORD p)            { /*###*/ };

    SIter_LL(const SortList_LL<T>& sl)      : C(sl) { i=C.h; };
    SIter_LL(const SIter_LL& cpy)           : C(cpy.C) { i=C.h; };
    //SIter_LL& operator=(const SIter_LL& )

    FRUITMACRO(SIter_LL<T>)
    friend class SortList_LL<T>;
  private:
    const SortList_LL<T>& C;
    DNode<T>*    i;

    void CheckI() const { if (!Valid()) Error(0); /* ### */ };
};
*/

/* /* /*

/* ------------------------------------------------------------------------ */
// Dictionary<K,V> & DIter<V>
// A dictionary contains two lists: One for storing a _Key and another for a
// _Value. You have to use Add(K,V) to add a pair of a key and a value which
// you will be able to access later by using Find(K) with a key. Find() will
// return an IterABC; use it to get back the Value with V Get(IterABC).
template<class K, class V> class Dictionary : public OCollectionABC<K, V>
{ public:
    void      Add(const K&, const &V);
    IterABC<K,V> Find(const I&) const;
    V&        Get(const IterABC<K, V>&);
    const V&  Get(const IterABC<K, V>&) const;
    void      Remove(const IterABC<L, V>&);

    void Add(const K&); // NOT defined! If used anyway, the linker detects it.
};

*/ */ */

/* ------------------------------------------------------------------------ */
// OCollUWI<OCollectionABC<I,O>&> - CollectionUserWorldInterface
// This FUNCTION (not method) converts ANY OrderedCollection into a string
// with format: {item1, item2, ...}
template<class I, class O>
String OCollUWI(const OCollectionABC<I,O>& C)
{ OIter<I, O> J=C.NewOIter();
  String s("{"); String i;
  while(J)
    { s+=J(); ++J; if(J) s+=", "; };
  s+="}";
  return s;
};
// Das obige soll eine allg. Prozedur zur Ausgabe von OCollectionABC's sein.
// Geht aber nicht, weil der dmliche BC nicht schnallt, dass ein Array<char>
// ein OCollectionABC<char,char> ist... (sollte er aber).
// Deshalb "von Hand" :
template<class T> String OCollUWI(const ArrayABC<T>& A)
{ typedef OCollectionABC<T,T>& rOC; return OCollUWI( TC(A,rOC) ); };
template<class T> String OCollUWI(const Array_A<T>& A)
{ typedef OCollectionABC<T,T>& rOC; return OCollUWI( TC(A,rOC) ); };
template<class T> String OCollUWI(const ListABC<T>& A)
{ typedef OCollectionABC<T,T>& rOC; return OCollUWI( TC(A,rOC) ); };
template<class T> String OCollUWI(const List_A<T>& A)
{ typedef OCollectionABC<T,T>& rOC; return OCollUWI( TC(A,rOC) ); };
template<class T> String OCollUWI(const SortListABC<T>& A)
{ typedef OCollectionABC<T,T>& rOC; return OCollUWI( TC(A,rOC) ); };
template<class T> String OCollUWI(const SortList_LL<T>& A)
{ typedef OCollectionABC<T,T>& rOC; return OCollUWI( TC(A,rOC) ); };
template<class T> String OCollUWI(const SortList_A<T>& A)
{ typedef OCollectionABC<T,T>& rOC; return OCollUWI( TC(A,rOC) ); };

/* ------------------------------------------------------------------------ */
// ContUWI<ItemT,CollectionT> - CollectionUserWorldInterface
// Derrive any Collection class from ContUWI and it can be outputet to cout!
/*
template<class T> class ContUWI : virtual public OCollectionABC<T,T>, public isASCIIable {
  String AsASCII() const
   { OIter<T,T,SWORD>& I=NewIter_OIter();
     String s("{");
     while(I)
       { s+=I(); I++; };
     s+="}";
     return s;             };
};
Das funktioniert aus div. Grnden (im Moment?) nicht wie gedacht...
*/

/* ------------------------------------------------------------------------ */
//                            IMPLEMENTATION
/* ------------------------------------------------------------------------ */
#pragma warn -rvl // turn off "Function should return a value"

template<class T> struct Node {
  T item;
  Node*  lnk;
  Node(T it, Node* n=NULL) : item(it), lnk(n) { };
};
/* ------------------------------------------------------------------------ */
template<class T> inline void Stack_LL<T>::
Add(const T& a)
{ Node<T>* N=new Node<T>(a);
  if(last)
    N->lnk=last;
  else
    head=N;
  last=N;
};

template<class T> inline T Stack_LL<T>::
GetRemove()
{ if(last) {
    T R=last->item;
    Node<T>* h=last->lnk;
    delete last;
    last=h;
    return R; }
  else
    Error(ErrContainerEmpty);  // should be throw
};

template<class T> inline void Stack_LL<T>::
~Stack_LL()
{};

/* ------------------------------------------------------------------------ */
template<class T> inline void Queue_LL<T>::
Add(const T& a)
{ Node<T>* N=new Node<T>(a);
  if(last)
    last->lnk=N;
  else
    head=N;
  last=N;
};

template<class T> inline T Queue_LL<T>::
GetRemove()
{ if(head) {
    T R=head->item;
    Node<T>* h=head->lnk;
    delete head;
    head=h;
    return R; }
  else
    Error(ErrContainerEmpty);  // should be throw
};

template<class T> inline Queue_LL<T>::
~Queue_LL()
{ while(head) {
    Node<T>* N=head->lnk;
    delete head;
    head=N;   };
};

/* ------------------------------------------------------------------------ */

template<class I, class O> inline O OCollectionABC<I,O>::
Get(const UWORD r) const
  { OIter<I,O> I=NewOIter(); I.Set(r); O& o=Get(I); return o; };

template<class I, class O> inline void OCollectionABC<I,O>::
Remove(const UWORD r)
  { OIter<I,O> I=NewOIter(); I.Set(r); Remove(I); };

template<class I, class O> inline O OCollectionABC<I,O>::
GetRemove()
  { O& o=Get(Items()); Remove(Items()); return o; };

/* ------------------------------------------------------------------------ */
template<class T> Array_A<T>::
Array_A(UWORD size) {
  if(size<1)
    Error(ErrArrayBounds);
  arr=new T[s=size];
  n=1;
};

template<class T> inline Array_A<T>::
Array_A(const Array_A<T>& cpy)
{ arr=new T[s=cpy.s];
  for(n=0; n<(cpy.n-1); n++)
    arr[n]=cpy.arr[n];
  n++;
};

template<class T> inline Array_A<T>& Array_A<T>::
operator= (const Array_A<T>& cpy)
  { if (cpy.s>s) Error(ErrArrayBounds);
    for (n=0; n<(cpy.n-1); n++)
      arr[n]=cpy.arr[n];
    n++;
    return *this;
  };

template<class T> inline void Array_A<T>::
Add(const T& i)
  { if(n>s) Error(ErrArrayOverflow);
    else arr[(n++)-1]=i;             };

template<class T> OIter<T,T> Array_A<T>::
Find_OIter(const T& c) const
  { for(UWORD q=0; q<(n-1); q++)
      if(arr[q]==c)
        return *new AIter_A<T>(*this,q+1);
//    AIter_A<T>& NotValid=*new AIter_A<T>(*this);
//    NotValid.i=0;    // Ungltiger Iterator!
//    return NotValid;
   return *new OIterErr<T,T>;
  };

template<class T> void Array_A<T>::
Remove(const UWORD r)
  { CheckBnd(r);
    for(UWORD q=r-1; q<(n-2); q++)
      arr[q]=arr[q+1];
    n--;
  };

template<class T> void Array_A<T>::
Remove(const IterABC<T,T>& I)
  { I.ValidFor(*this); I.Valid();                // Wenn I ValidFor dieses Array
    for(UWORD q=TC(I,AIter_A<T>&).i-1; q<(n-2); q++) // ist, ist es sicher ein AIter_A.
      arr[q]=arr[q+1];
    n--;
  };

/* ------------------------------------------------------------------------ */

/*
template<class T> void ListABC<T>::
Insert(const UWORD r, const T it)
  { OIter<I,O>& I=NewOIter(); I.Set(r); Insert(I,it); delete &I; };
*/
/* ------------------------------------------------------------------------ */

template<class T> void List_A<T>::
Insert(const IterABC<T,T>& I, const T it)
  { I.ValidFor(A); I.Valid();
    UWORD p=TC(I,AIter_A<T>&).i; // siehe oben
    if(A.n>A.s) {
      T* old=A.arr;
      A.arr=new T[A.s+=incr];
      { UWORD i;
        for(i=0; i<(p-1); i++)
          A.arr[i]=old[i];
        for(i=A.n; i>(p-1); i--)
          A.arr[i]=old[i-1];     };
      delete[] old;
    }
    else {
      for(UWORD i=A.n; i>(p-1); i--)
        A.arr[i]=A.arr[i-1];
      A.n++;
    };
    A.arr[p-1]=it;
};

//template<class T> void List_A<T>::
//Insert(const UWORD i, const T it)

template<class T> void List_A<T>::
Add(const T& it)
{
  if(A.n>A.s) {
    T* old=A.arr;
    A.arr=new T[A.s+=incr];
    for(UWORD i=0; i<(A.n-1); i++)
      A.arr[i]=old[i];
    delete[] old;
  };
  A.arr[(A.n++)-1]=it;
};

/* ------------------------------------------------------------------------ */

template<class T> void SortList_A<T>::
Add(const T& e)
{
  UWORD i=1;
  if(i<=A.Items()) {
    while(e>A.Get(i))
      { i++; if(i>A.Items()) break; };
  }

  if(i>A.Items()) A.Add(e);
  else            A.Insert(i,e);
};

/* ------------------------------------------------------------------------ */

/* SortList_LL - a sorted listed implemented with a double linked list
   DOES NOT WORK! PLEASE FINISH LATER!

template<class T> struct DNode {
  T item;
  DNode* n; DNode* p;
  DNode(T it, DNode* na=NULL, DNode* pa=NULL) : item(it), n(na), p(pa) { };
};

template<class T> inline void SortList_LL<T>::
Add(const T& i)
{ DNode<T>* lc=h; DNode<T>* c=h;
  while(c) {
    lc=c;
    if(i>(c->item))  c=c->n;
    else           c=NULL;    };
  if(h) { DNode<T>* N=new DNode<T>(i,lc,lc->p);  lc->p->n=N; lc->p=N; }
  else  { h=new DNode<T>(i,NULL,NULL);                         };
  ++l;
};

template<class T> SortList_LL<T>::~SortList_LL()
 { DNode<T>* lc=h; DNode<T>* N;
   while(lc)  {
     N=lc->n;
     delete lc;
     lc=N;     };
   h=NULL;
 };
*/

/* ------------------------------------------------------------------------ */
/* ------------------------------------------------------------------------ */
